/* server.c
 *
 * Simulation control functions used only by the server
 */
#include "sim.h"
#include "server.h"

static void send_data_directory(int fd, char *datadir, int trmode, char *name);
static void dump_data(char *fn, void (*wfunc)(int fd));
static void drop_item(person *pr, object *ob);
static void backup_file(char *fn);
static int count_thing(void (*reset)(void),void *(*next)(void));

extern int dont_dump;

/* process_request
 * Read a request from a player and act upon it. See protocol.html for 
 * request format.
 */
void process_request(player *p)
{
person *pr;			/* person representing this player */
char reqline[256];		/* request line from client */
char **reqv;			/* tokens from reqline */
int reqc;			/* number of tokens in reqv */
static struct {			/* expected no. of args for requests */
	int rnum;
	int argc;
	} ar[] = {req_move,2, req_say,-1, req_yell,-1, req_get,2,
		  req_do,-1, req_drop,2, req_turn,2, req_give,3, 
		  req_upload,4, req_send,3, req_actlist,2, req_ls,1,
		  req_rm,2, req_setpass,2, req_suicide,1, req_avatar,1,
		  req_email,2, req_fullname,2 };
int rnum;			/* request number */
int i;

pr = lookup_person(p->name);

if (fdgets(p->fd,reqline,256) == NULL) {
	/* Connection closed.
	 */
	disconnect_player(p,1);
	return;
	}

/* Parse and check command line
 */
parse_string(reqline,&reqv,&reqc);
if (reqc == 0) {
	logerror("Null command received from %s",p->name);
	return;
	}
rnum = atoi(reqv[0]);
for(i=0; i<sizeof(ar)/sizeof(ar[0]); i++)
	if (ar[i].rnum == rnum && ar[i].argc != -1 && ar[i].argc != reqc) {
		logerror("Wrong number of args for reqest %d from %s",
			 rnum,p->name);
		return;
		}

/* Execute
 */
switch(atoi(reqv[0])) {
case req_move: {
	int ln = atoi(reqv[1]);		/* link number taken */
	path *pt;			/* path taken */
	node *nn;			/* new position */
	shape *sh;			/* shape representing player */
	person *opr;			/* someone else blocking way */

	/* Check if node exists and path is open
	 */
	if (ln < 0 || ln >= pr->pos->linkc) {
		server_command(p->fd,com_err,err_nopath);
		return;
		}
	pt = pr->pos->link[ln];
	if (pt->closed) {
		server_command(p->fd,com_err,err_blocked);
		return;
		}

	/* Check if no-one else is there
	 */
	nn = other_end(pt,pr->pos);

	/* Make the move
	 */
	person_move(pr,nn);
	break;
	}

case req_say: {
	player *pp;
	person *ppr, *npc;
	char *msgv[256];
	int i;

	/* Send message to players nearby
	 */
	msgv[0] = p->name;
	msgv[1] = ":";
	for(i=1; i<reqc; i++)
		msgv[i+1] = reqv[i];
	reset_player();
	while((pp = next_online_player())) {
		ppr = lookup_person(pp->name);
		if (close_node(ppr->pos,pr->pos)) {
			server_command(pp->fd,com_msg,msgv,reqc+1);
			}
		}

	/* Send message to npc persons nearby
	 */
	reset_person();
	while((npc = next_person())) {
		action *ac = lookup_action(npc->name);
		if (ac && close_node(npc->pos,pr->pos))
			exec_action(ac,p,"listen",msgv+2,reqc-1);
		}
	break;
	}

case req_yell: {
	player *pp;
	person *npc;
	char *msgv[256];
	int i;

	/* Send message to players nearby
	 */
	msgv[0] = p->name;
	msgv[1] = "yells";
	msgv[2] = ":";
	for(i=1; i<reqc; i++)
		msgv[i+2] = reqv[i];
	server_command_all(com_msg,msgv,reqc+2);

	/* Send message to npc persons nearby
	 */
	reset_person();
	while((npc = next_person())) {
		action *ac = lookup_action(npc->name);
		if (ac && close_node(npc->pos,pr->pos))
			exec_action(ac,NULL,"listen",msgv+3,reqc-1);
		}
	break;
	}

case req_get: {
	object *ob = lookup_object(reqv[1]);

	if (!ob || ob->pos != pr->pos)
		server_command(p->fd,com_err,err_notfound);
	else if (pr->invc >= INV_SIZE) 
		server_command(p->fd,com_err,err_invfull);
	else if (!ob->gettable)
		server_command(p->fd,com_err,err_cantget);
	else if (player_data_size(p) + object_data_size(ob) > MAX_PLAYER_DATA)
		server_command(p->fd,com_err,err_maxdata);
	else {
		shape *sh = lookup_shape(ob->name);

		pickup_item(pr,ob);
		server_command_all(com_getobject,ob->name,pr->name);
		server_command_all(com_delshape,sh->name);
		sh->hidden = 1;
		}
	break;
	}

case req_do: {
	object *ob;
	action *ac;

	if (reqc < 3)
		return;
	ob = lookup_object(reqv[2]);
	if (!ob || (ob->inv != pr && !close_node(ob->pos,pr->pos)))
		server_command(p->fd,com_err,err_notfound);
	else if (reqv[1][0] == '_' || !(ac = lookup_action(ob->name)) ||
		 !exec_action(ac,p,reqv[1],&reqv[3],reqc-3))
		server_command(p->fd,com_err,err_noaction);
	break;
	}

case req_drop: {
	object *ob = lookup_object(reqv[1]);

	if (!ob || ob->inv != pr)
		server_command(p->fd,com_err,err_donthave);
	else
		drop_item(pr,ob);
	break;
	}

case req_turn: {
	double ang = atof(reqv[1]);
	shape *sh = lookup_shape(p->name);
	player *pp;

	pr->ang = ang;
	sh->rota = ang; sh->rotx = 0; sh->roty = 0; sh->rotz = 1.0;
	reset_player();
	while((pp = next_online_player()))
		if (pp->ilevel != lvl_none)
			server_command(pp->fd,com_rotshape,sh->name,ang,
				       0.0,0.0,1.0);
	break;
	}

case req_give: {
	object *ob = lookup_object(reqv[1]);
	person *rpr = lookup_person(reqv[2]);

	if (!ob || ob->inv != pr)
		server_command(p->fd,com_err,err_notfound);
	else if (!rpr || !close_node(pr->pos,rpr->pos))
		server_command(p->fd,com_err,err_nothere);
	else {
		putdown_item(pr,ob);
		pickup_item(rpr,ob);
		server_command_all(com_giveobject,ob->name,pr->name,rpr->name);
		}
	break;
	}

case req_upload: {
	int size = atoi(reqv[2]);	/* size of file player is uploading */
	char filbuf[256];
	int f;				/* fd for writing to data file */

	if (player_data_size(p) + size > MAX_PLAYER_DATA) {
		/* This file would exceed max data per player
		 */
		server_command(p->fd,com_err,err_maxdata);
		break;
		}
	sprintf(filbuf,"%s/player_data/%s/%s",mp->mapdir,p->name,reqv[1]);
	if ((f = open_data_file(filbuf)) == -1) {
		logerror("Error creating player data file %s",filbuf);
		return;
		}
	if (!sfork()) {
		/* Receive data to file concurrently
		 */
		int rv;

		set_signals(SIG_DFL);
		rv = recv_data(p->ip,atoi(reqv[3]),f,size);
		close(f);
		if (rv == 0) {
			logerror("Error reading data from player %s",p->name);
			unlink(filbuf);
			}
		exit(0);
		}
	close(f);
	break;
	}

case req_send: {
	char datafile[256];		/* player's data file */
	player *rpr = lookup_player(reqv[2]);
	int f;				/* fd for datafile */

	sprintf(datafile,"%s/player_data/%s/%s",mp->mapdir,p->name,reqv[1]);
	if (!rpr || !close_node(pr->pos,lookup_person(rpr->name)->pos))
		server_command(p->fd,com_err,err_nothere);
	else if (rpr->fd == -1)
		server_command(p->fd,com_err,err_notonline);
	else if (send_mime_data(rpr->fd,datafile) == 0)
		server_command(p->fd,com_err,err_nofile);
	break;
	}

case req_actlist: {
	object *obj = lookup_object(reqv[1]);
	char *acts[256];
	int actc = 0;

	if (!obj || (obj->inv != pr && !close_node(obj->pos,pr->pos)))
		server_command(p->fd,com_err,err_notfound);
	else {
		action *ac = lookup_action(obj->name);

		if (ac != NULL) {
			int i;

			for(i=0; i<ac->metc; i++) {
				char *mn = ac->met[i]->name[0];

				if (mn[0] != '_')
					acts[actc++] = mn;
				}
			}
		server_command(p->fd,com_actlist,obj->name,acts,actc);
		}
	break;
	}

case req_ls: {
	char *dats[256];	/* array of filenames */
	int datc = 0;
	char datadir[256];	/* player's data dir */
	struct stat stbuf;	/* for checking if dir exists */
	DIR *dlist;		/* for reading dir */
	struct dirent *dent;	/* info on each dir entry */
	int i;

	sprintf(datadir,"%s/player_data/%s",mp->mapdir,p->name);
	if ((dlist = opendir(datadir)) == NULL) {
		/* no data dir associated with this player
		 */
		server_command(p->fd,com_ls,NULL,0);
		break;
		}
	while((dent = readdir(dlist))) {
		if (dent->d_ino && dent->d_name[0] != '.') {
			dats[datc++] = strdupe(dent->d_name);
			}
		}
	closedir(dlist);
	server_command(p->fd,com_ls,dats,datc);
	for(i=0; i<datc; i++)
		free(dats[i]);
	break;
	}

case req_rm: {
	char *base = basenam(reqv[1]);
	char fullpath[256];

	sprintf(fullpath,"%s/player_data/%s/%s",mp->mapdir,p->name,base);
	if (strequ(base,".") || strequ(base,".."))
		server_command(p->fd,com_err,err_nofile);
	else if (unlink(fullpath))
		server_command(p->fd,com_err,err_nofile);
	break;
	}

case req_setpass: {
	if (p->password) free(p->password);
	if (reqv[1][0]) p->password = strdupe(reqv[1]);
	else            p->password = NULL;
	break;
	}

case req_suicide: {
	int i;
	server_message(p->fd,"Thank you for living..");
	server_command(p->fd,com_shutdown);
	disconnect_player(p,0);
	for(i=pr->invc-1; i>=0; i--)
		drop_item(pr,lookup_object(pr->inv[i]));
	delete_player(p);
	break;
	}

case req_avatar: {
	char *tf = store_to_end(p->fd);
	int afd = open(tf,O_RDONLY);
	shape *nsh;			/* new avatar */
	char *why;

	if (parse_shape(afd,&nsh) != 0)
		server_command(p->fd,com_err,err_parse);
	else if ((why = avatar_sanity_check(nsh))) {
		server_command(p->fd,com_err,err_avatar);
		server_message(p->fd,"%s",why);
		free_shape(nsh);
		}
	else {
		shape *osh = lookup_shape(p->name);
		player *pp;

		/* Remove old shape on clients and server
		 */
		server_command_all(com_delshape,osh->name);
		remove_shape(osh);
		free_shape(osh);

		/* Add new shape to server and clients
		 */
		unname_shape(nsh);
		strcpy(nsh->name,p->name);
		nsh->rotx = 0; nsh->roty = 0; nsh->rotz = 1;
		nsh->rota = pr->ang;
		move_shape(nsh,pr->pos->x,pr->pos->y,pr->pos->z);
		add_shape(nsh);
		reset_player();
		while((pp = next_online_player())) {
			if (pp->ilevel != lvl_none) {
				server_command(pp->fd,com_addshape,nsh->name);
				write_shape(pp->fd,0,nsh,pp->ilevel,0);
				}
			}
		}
	close(afd);
	unlink(tf);
	break;
	}

case req_email: {
	if (pr->email) free(pr->email);
	pr->email = strdupe(reqv[1]);
	server_command_all(com_email,p->name,reqv[1]);
	break;
	}

case req_fullname: {
	if (pr->fullname) free(pr->fullname);
	pr->fullname = strdupe(reqv[1]);
	server_command_all(com_fullname,p->name,reqv[1]);
	break;
	}

default:
	logerror("Unknown request number %d from %s",rnum,p->name);
	return;

	} /* end switch */
}


/* control_others
 * Move persons and objects not controlled by players
 */
void control_others(void)
{
person *pr;
reset_person();
while((pr = next_person())) {
	action *ac = lookup_action(pr->name);

	if (ac) {
		variable *tt = lookup_variable("thinktime",ac,NULL);

		if (time(NULL) - ac->lastthought > (tt ? tt->val.dbl : 2.0)) {
			exec_action(ac,NULL,"think",NULL,0);
			ac->lastthought = time(NULL);
			}
		}
	}
}


/* dump_state
 * Dump the current server state to files. The original map file is overwritten
 * and new nodes, paths, etc.. files are created containing the current state.
 */
void dump_state(void)
{
char mfile[256];	/* full path to map file */
int fd;			/* fd for map file */

if (dont_dump) return;
/* logmessage("Dumping server state to dir %s",mp->mapdir); */

/* Write out map file
 */
sprintf(mfile,"%s/%s",mp->mapdir,mp->mapfile);
backup_file(mfile);
if ((fd = creat(mfile,0700)) == -1) {
	logerror("Error creating map file %s",mfile);
	return;
	}
write_map(fd,0);
close(fd);

/* Write out all the other data files
 */
dump_data(mp->acfile,write_actions);
dump_data(mp->ndfile,write_nodes);
dump_data(mp->obfile,write_objects);
dump_data(mp->ptfile,write_paths);
dump_data(mp->prfile,write_persons);
dump_data(mp->plfile,write_players);
dump_data(mp->shfile,write_all_shapes);

/* logmessage("Done dumping to dir %s",mp->mapdir); */
}


/* dump_data
 * Output nodes, paths, etc.. data to a named file
 */
static void dump_data(char *fn, void (*wfunc)(int fd))
{
char filename[256];
int fd;

sprintf(filename,"%s/%s",mp->mapdir,fn);
backup_file(filename);
if ((fd = creat(filename,0700)) == -1) {
	logerror("Error creating map data file %s",filename);
	return;
	}
wfunc(fd);
close(fd);
}


/* backup_file
 * Move all data from the file fn to the file fn.backup
 */
static void backup_file(char *fn)
{
char bfn[256];

sprintf(bfn,"%s.backup",fn);
if (rename(fn,bfn) == -1)
	logerror("Failed to backup file %s to %s!",fn,bfn);
}


/* disconnect_players
 * Tell all connected players to log off now.
 */
int disconnect_players(void)
{
player *pp;
int count = 0;

reset_player();
while((pp = next_player())) {
	if (pp->fd != -1) {
		server_message(pp->fd,"Server shutting down! Disconnecting.");
		server_command(pp->fd,com_shutdown);
		count++;
		}
	disconnect_player(pp,0);
	}
return count;
}


/* delete_player
 * Remove a player, their shape and person from the server. Also delete any
 * objects they are carrying and data belonging to this player.
 * Tell other clients about it.
 * Note - assumes this player is already disconnected
 */
void delete_player(player *p)
{
char datadir[256];
int i;

/* Delete player data
 */
sprintf(datadir,"%s/player_data/%s",mp->mapdir,p->name);
delete_dir(datadir);

delete_person(lookup_person(p->name));

/* Remove player, person and shape
 */
remove_player(p);
free_player(p);
}


/* delete_person
 * Delete a person, their action and all their objects
 */
void delete_person(person *pr)
{
int i;
shape *sh;
action *ac;

/* Remove objects, their shapes and data
 */
for(i=0; i<pr->invc; i++) {
	object *iob = lookup_object(pr->inv[i]);
	shape  *ish = lookup_shape(pr->inv[i]);
	action *iac = lookup_action(pr->inv[i]);
	char datadir[256];

	sprintf(datadir,"%s/object_data/%s",mp->mapdir,iob->name);
	delete_dir(datadir);
	server_command_all(com_delobject,iob->name);
	remove_object(iob);
	free_object(iob);
	remove_shape(ish);
	free_shape(ish);
	if (iac) {
		remove_action(iac);
		free_action(iac);
		}
	}

/* Remove person, shape and action
 */
ac = lookup_action(pr->name);
sh = lookup_shape(pr->name);
server_command_all(com_delperson,pr->name);
server_command_all(com_delshape,pr->name);
remove_person(pr);
free_person(pr);
remove_shape(sh);
free_shape(sh);
if (ac) {
	remove_action(ac);
	free_action(ac);
	}
}


/* init_data_dir
 * Make sure the named dir in the map dir exists
 */
void init_data_dir(char *dir)
{
char fullpath[256];
struct stat stbuf;

sprintf(fullpath,"%s/%s",mp->mapdir,dir);
if (stat(fullpath,&stbuf) == -1) {
	/* dir doesn't exist at all
	 */
	mkdir(fullpath,0700);
	}
else if ((stbuf.st_mode & S_IFMT) != S_IFDIR) {
	/* A file is in the way.
	 */
	if (unlink(fullpath) == -1) {
		logerror("Couldn't remove file %s",fullpath);
		exit(5);
		}
	mkdir(fullpath,0700);
	}
}


/* person_move
 * Move a person (who may be a player) to a new node, which may be on
 * another server
 */
void person_move(person *pr, node *nn)
{
player *p = lookup_player(pr->name);
shape *sh  = lookup_shape(pr->name);
player *pp;

if (!nn->special) {
	/* Move within this server
	 */
	double rx = nn->x - pr->pos->x;
	double ry = nn->y - pr->pos->y;
	double rz = nn->z - pr->pos->z;
	object *ob;

	/* Move player
	 */
	pr->pos = nn;
	if (p && p->fd != -1)
		server_command(p->fd,com_move,pr->pos->name);

	/* Tell all clients to move the person and shape representing
	 * this player.
	 */
	server_command_all(com_moveperson,pr->name,pr->pos->name);
	move_shape(sh,rx,ry,rz);
	reset_player();
	while((pp = next_online_player()))
		if (pp->ilevel != lvl_none)
			server_command(pp->fd,com_moveshape,sh->name,rx,ry,rz);

	/* Check for objects at the person's location with the _trigger
	 * method defined for their action.
	 */
	reset_object();
	while((ob = next_object())) {
		action *ac = lookup_action(ob->name);

		if (ob->pos == pr->pos && ac)
			exec_action(ac,p,"_trigger",NULL,0);
		}
	}
else {
	int sfd;		/* connection to other server */
	char datadir[256];	/* name of dir of object/player data */
	int i;

	/* Connect to other server.
	 */
	if (p)
		logmessage("Transferring player %s to %s port %d",
			   p->name,nn->serv.addr,nn->serv.port);
	sfd = open_tcp_send_ip(nn->serv.ip,nn->serv.port);
	if (sfd == -1) {
		logerror("Couldn't connect to %s port %d for transfer",
			 nn->serv.addr,nn->serv.port);
		if (p && p->fd != -1)
			server_command(p->fd,com_err,err_down);
		return;
		}
	if (!send_version(sfd)) {
		if (p && p->fd != -1)
			server_command(p->fd,com_err,err_down);
		return;
		}
	
	/* Move person to new position, and shape to 0,0,0
	 */
	move_shape(sh,-(pr->pos->x),-(pr->pos->y),-(pr->pos->z));
	pr->pos = nn;

	if (p) {
		/* Send player struct
		 */
		fdprintf(sfd,"%d\n",con_transfer_player);
		if (!send_password(sfd,mp->password)) {
			if (p->fd != -1) server_command(p->fd,com_err,err_down);
			return;
			}
		write_player(sfd,0,p);
		}
	
	/* Send person
	 */
	fdprintf(sfd,"%d\n",con_transfer_person);
	send_password(sfd,mp->password);
	write_person(sfd,0,pr);
	
	/* Send person's shape
	 */
	fdprintf(sfd,"%d\n",con_transfer_shape);
	send_password(sfd,mp->password);
	write_shape(sfd,0,sh,lvl_textured,1);
	
	/* Send person's objects, their actions, shapes and data
	 */
	for(i=0; i<pr->invc; i++) {
		object *ob  = lookup_object(pr->inv[i]);
		action *oac = lookup_action(ob->name);
		shape  *osh = lookup_shape(ob->name);
	
		/* Send object
		 */
		fdprintf(sfd,"%d\n",con_transfer_object);
		send_password(sfd,mp->password);
		write_object(sfd,0,ob);
	
		/* Send action
		 */
		if (oac) {
			fdprintf(sfd,"%d\n",con_transfer_action);
			send_password(sfd,mp->password);
			write_action(sfd,0,oac);
			}

		/* Send shape
		 */
		fdprintf(sfd,"%d\n",con_transfer_shape);
		send_password(sfd,mp->password);
		write_shape(sfd,0,osh,lvl_textured,1);

		/* Send data
		 */
		sprintf(datadir,"%s/object_data/%s",mp->mapdir,ob->name);
		send_data_directory(sfd,datadir,con_transfer_obdata,ob->name);
		}

	if (p) {
		/* Send player's data
		 */
		sprintf(datadir,"%s/player_data/%s",mp->mapdir,p->name);
		send_data_directory(sfd,datadir,con_transfer_pldata,p->name);
		}

	/* Disconnect from other server
	 */
	close(sfd);
	
	if (p) {
		/* Tell this player where to go
		 */
		if (p->fd != -1)
			server_command(p->fd,com_newserv,nn->serv.addr,
							 nn->serv.port);
	
		/* Remove the player from this server
		 */
		disconnect_player(p,1);
		delete_player(p);
		}
	else
		delete_person(pr);
	}
}


/* send_data_directory
 * Send a whole directory of object or player data to another server
 */
static void send_data_directory(int fd, char *datadir, int trmode, char *name)
{
DIR *dlist;		/* for reading dir */
struct dirent *dent;	/* info on each dir entry */

if ((dlist = opendir(datadir))) {
	while((dent = readdir(dlist))) {
		if (dent->d_ino && dent->d_name[0] != '.') {
			char *fname = dent->d_name;	/* file name */
			char datafil[256];		/* file path */
			struct stat stbuf;		/* info on file */
			int port = mp->port + alloc_port();
			int f;

			sprintf(datafil,"%s/%s",datadir,fname);
			if (stat(datafil,&stbuf) == -1) {
				logerror("Couldn't open file %s for transfer ",
					 "to other sever",datafil);
				continue;
				}

			/* Send transfer info to other server
			 */
			fdprintf(fd,"%d\n",trmode);
			send_password(fd,mp->password);
			fdprintf(fd,"%s\n",name);
			fdprintf(fd,"%s\n",fname);
			fdprintf(fd,"%d\n",port);
			fdprintf(fd,"%d\n",stbuf.st_size);

			/* Start concurrent transfer of actual data
			 */
			f = open(datafil,O_RDONLY);
			if (!sfork()) {
				set_signals(SIG_DFL);
				if (send_data(port,f,stbuf.st_size) == 0)
					logerror("Error transferring file %s "
						 "to other server",datafil);
				close(f);
				exit(0);
				}
			close(f);
			}
		}
	closedir(dlist);
	}
}


/* disconnect_player
 * Disconnect a player from this server
 */
void disconnect_player(player *p, int tellothers)
{
if (p->fd != -1) {
	shape *sh = lookup_shape(p->name);

	close(p->fd);
	p->fd = -1;
	p->inout = time(NULL);
	shade_shape(sh,sft_glass);
	if (tellothers)
		server_command_all(com_shdshape,sh->name,sft_glass);
	logmessage("Player %s has left the map",p->name);
	}
}


/* drop_item
 * Drop an object belonging to a person
 */
static void drop_item(person *pr, object *ob)
{
shape *sh = lookup_shape(ob->name);
player *pp;
double nx = pr->pos->x + (nrand()-0.5) / 2;
double ny = pr->pos->y + (nrand()-0.5) / 2;
double nz = pr->pos->z;

putdown_item(pr,ob);
server_command_all(com_dropobject,ob->name,pr->name);
sh->hidden = 0;
move_shape(sh,nx - sh->pos.x, ny - sh->pos.y, nz - sh->pos.z);
reset_player();
while((pp = next_online_player())) {
	if (pp->ilevel != lvl_none) {
		server_command(pp->fd,com_addshape,sh->name);
		write_shape(pp->fd,0,sh,pp->ilevel,0);
		}
	}
}


/* player_data_size
 * Returns the total size of all data associated with a player or with 
 * objects carried.
 */
long player_data_size(player *p)
{
long c;			/* total bytes */
char datadir[256];	/* player's data dir name */
person *pr;
int i;

/* Count up player's files
 */
sprintf(datadir,"%s/player_data/%s",mp->mapdir,p->name);
c = directory_size(datadir);

/* Count up carried object's data
 */
pr = lookup_person(p->name);
for(i=0; i<pr->invc; i++) {
	object *ob = lookup_object(pr->inv[i]);
	c += object_data_size(ob);
	}

return c;
}

/* object_data_size
 * Returns the total size of all data associated with an object
 */
long object_data_size(object *o)
{
char datadir[256];

sprintf(datadir,"%s/object_data/%s",mp->mapdir,o->name);
return directory_size(datadir);
}


/* check_old_players
 * Check every player for >5 days inactivity, and if so make them invisible
 */
void check_old_players(void)
{
player *p;
time_t now = time(NULL);

reset_player();
while((p = next_player()) != NULL) {
	shape *sh = lookup_shape(p->name);

	if (p->fd == -1 && !sh->hidden && now - p->inout > FADEOUT_TIME) {
		sh->hidden = 1;
		server_command_all(com_delshape,sh->name);
		}
	}
}


/* avatar_sanity_check
 * Returns NULL if an avatar is valid, or a string saying why
 */
char *avatar_sanity_check(shape *sh)
{
double bx1,by1,bz1, bx2,by2,bz2;
int comp;
static char why[256];

comp = shape_box(sh,&bx1,&by1,&bz1,&bx2,&by2,&bz2);
if (comp < 2) {
	sprintf(why,"Avatar consists of only %d shapes. Minimum is 2.",comp);
	return why;
	}
if (comp > 20) {
	sprintf(why,"Avatar consists of %d shapes. Maximum is 20.",comp);
	return why;
	}
if (bx1 > 0 || by1 > 0 || bx2 < 0 || by2 < 0 || bz2 < 0) {
	return "Avatar is not positioned about 0,0,0.";
	}
if (bx2 - bx1 > 3) {
	sprintf(why,"Avatar is %f metres wide in the x direction. Maximum is 3",
		    bx2 - bx1);
	return why;
	}
if (by2 - by1 > 3) {
	sprintf(why,"Avatar is %f metres wide in the y direction. Maximum is 3",
		    by2 - by1);
	return why;
	}
if (bz2 - bz1 > 5) {
	sprintf(why,"Avatar is %f metres high in the z direction. Maximum is 5",
		    bz2 - bz1);
	return why;
	}
if (bz2 > 5) {
	sprintf(why,"Avatar is %f metres above the ground. Maximum is 5",bz2);
	return why;
	}
return NULL;
}


/* count_thing
 * Generic macro for counting objects / persons / etc..
 */
#define count_thing(thing,c) \
{ \
	*(c) = 0; \
	cat2(reset_,thing)(); \
	while(cat2(next_,thing)()) (*(c))++; \
}


/* contact_metaserver
 * Connect to the metaserver, a cgi program on a web server which records the
 * state of running sim servers. Command line args passed to the cgi program
 * are :  host port players person objects shapes actions nodes paths
 */
void contact_metaserver(void)
{
if (!sfork()) {
	int fd = open_tcp_send(METASERVER_HOST,METASERVER_PORT);
	int plc, prc, obc, shc, acc, ndc, ptc;

	if (fd == -1)
		exit(1);
	count_thing(player,&plc);
	count_thing(person,&prc);
	count_thing(object,&obc);
	count_thing(shape,&shc);
	count_thing(action,&acc);
	count_thing(node,&ndc);
	count_thing(path,&ptc);
	fdprintf(fd,"GET /cgi-bin/metaserver?%s+%d+%d+%d+%d+%d+%d+%d+%d "
		    "HTTP/1.0\n\n",
		 mp->server,mp->port,plc,prc,obc,shc,acc,ndc,ptc);
	close(fd);
	exit(0);
	}
}


